# -*- coding: utf-8 -*-
# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

__author__ = "FMR LLC"
__email__ = "opensource@fmr.com"
__version__ = "1.1.0"
__copyright__ = "Copyright (C), FMR LLC"
